image.symbolic-circular {
  background-color: alpha(currentColor, 0.08);
  min-width: 32px;
  min-height: 32px;
  border-radius: 50%;
}

.warning-box image,
.warning-box label {
  color: @error_color;
}

.warning-box label {
  font-weight: 600;
}
